/**************************************************************************
 * 
 * PROJECT: Task Control Management (TCM)
 *
 * (c) Copyright 1998 Reid Simmons.  All rights reserved.
 *
 * FILE: tcmNddsResponder.cc
 *
 * ABSTRACT: Functions for responding to TCM invocations from ControlShell,
 *           via NDDS
 *
 * EXPORTS:
 *
 * $Revision: 1.3 $
 * $Date: 1998/04/21 17:58:21 $
 * $Author: reids $
 *
 * REVISION HISTORY:
 * $Log: tcmNddsResponder.cc,v $
 * Revision 1.3  1998/04/21 17:58:21  reids
 * Changes to make stimulus/response messages be reliable productions in NDDS
 *
// Revision 1.2  98/03/18  14:41:50  reids
// Changes to support both SunOS and Solaris versions.
// 
// Revision 1.1  98/01/02  00:21:11  reids
// Initial attempt to integrate TCM and ControlShell via NDDS.
//   A bit kludgy, since I don't really understand how to handle a series
//   of productions -- really want them to be queued, reliably...
// 
 **************************************************************************/

#include "cs.h"
#include "CSFsm.h"
#include "CSFsmResponseModule.pkg.h"
#include "tcmNddsResponder.h"
#ifdef __svr4__
extern "C" int sleep (unsigned int);
#endif
#ifndef NDDS_FIFO_DEPTH_NO_CHANGE // Hack: macro found in NDDS 1.11 but not 1.9
#define NDDS_1_9
#endif

static NDDSConsumer Response_Consumer = NULL;
static NDDSProducer Response_Producer = NULL;

static NDDSConsumer Stimulus_Consumer = NULL;
static NDDSProducer Stimulus_Producer = NULL;

static TCM_NDDS_Response_Type tcmNddsResponseVar;
static TCM_NDDS_Stimulus_Type tcmNddsStimulusVar;

static CSFsm _fsm = NULL;

#define TRACE_RESPONDER

/**************************************************************************
 *
 * CLASS: tcmNddsResponderClass
 *
 **************************************************************************/

boolean tcmNddsResponderClass::respond (String context, int returnCode) const
{
  String responseString; 

  responseString = _responseTable.getResponse((String)context, returnCode);
  if (responseString) {
#ifdef TRACE_RESPONDER
    fprintf(stderr, "tcmNddsResponderClass:respond: %s (%d) => %d/%s\n", 
	    context, returnCode, _refId,
	    (streq(responseString, SUCCESS_RESPONSE)
	     ? "SUCCESS" : responseString));
#endif
    tcmNddsRespond(_refId, responseString);
  }
  return (responseString != NULL);
}

// Respond in case the current stimulus was not handled by the current state.

boolean tcmNddsResponderClass::missingStimulus (String stimulus) const
{
  boolean sameResponder;

  sameResponder = !strcmp(stimulus, _stimulus);
  if (sameResponder) {
#ifdef TRACE_RESPONDER
    fprintf(stderr, "tcmNddsResponderClass:missingStimulus: %s\n", stimulus);
#endif
    tcmNddsRespond(_refId, NO_STIM_MODULE);
  }
  return sameResponder;
}

/****************************************************************
 *
 * Internal Functions
 *
 ****************************************************************/

/**************************************************************************
 *
 * FUNCTION: NDDSObjectInstance tcmNddsResponseHandler (NDDSUpdateInfo ui)
 *
 * DESCRIPTION: Received a response from CS -- handle it.
 *
 *******************************************************************/
static NDDSObjectInstance tcmNddsResponseHandler (NDDSUpdateInfo ui)
{
  if (ui->instance && (ui->updateStatus == NDDS_FRESH_DATA)) {
    TCM_NDDS_Response_Ptr response = (TCM_NDDS_Response_Ptr)(ui->instance);
    
#ifdef TRACE_RESPONDER
    fprintf(stderr, "tcmNddsResponseHandler: id: %d, response: %s\n", 
	    response->id, response->response);
#endif
    if (!response->response || streq(response->response, SUCCESS_RESPONSE)) {
      TCM_Success(TCM_Retrieve(response->id));
    } else {
      TCM_Failure(TCM_Retrieve(response->id), response->response);
    }
  }
  return ui->instance;
}

/**************************************************************************
 *
 * FUNCTION: NDDSObjectInstance tcmNddsStimulusHandler (NDDSUpdateInfo ui)
 *
 * DESCRIPTION: Received a stimulus from TCM -- handle it.
 *
 *******************************************************************/
static NDDSObjectInstance tcmNddsStimulusHandler (NDDSUpdateInfo ui)
{
  if (ui->instance && (ui->updateStatus == NDDS_FRESH_DATA)) {
    TCM_NDDS_Stimulus_Ptr stimulus = (TCM_NDDS_Stimulus_Ptr)(ui->instance);

#ifdef TRACE_RESPONDER
    fprintf(stderr, "tcmNddsStimulusHandler: id: %d, stimulus: %s\n", 
	    stimulus->id, stimulus->stimulus);
#endif
    tcmNddsResponder responder
      = new tcmNddsResponderClass(stimulus->stimulus, stimulus->id,
				  stimulus->numResponses);
    responder->add(_fsm);
    for (int i=0; i < stimulus->numResponses; i++) {
      responder->addResponse(stimulus->responses[i].context,
			     stimulus->responses[i].returnCode,
			     stimulus->responses[i].responseString);
    }

    // Send stimulus
    _fsm->StimulusSend(stimulus->stimulus, 0);
  }
  return ui->instance;
}

static void consumerReliableStatusFn (NDDSConsumerReliableStatus status)
{
  if (status->event != NDDS_PRODUCER_NEW) {
    fprintf(stderr, "WARNING: Problem with reliable consumer for %s: %d\n",
	    status->nddsName, status->event);
  }
}

static void producerReliableStatusFn (NDDSProducerReliableStatus status)
{
  if (status->event != NDDS_SUBSCRIBER_NEW) {
    fprintf(stderr, "WARNING: Problem with reliable producer for %s: %d\n",
	    status->nddsName, status->event);
  }
}

/****************************************************************
 *
 * Public Functions
 *
 ****************************************************************/

/**************************************************************************
 *
 * FUNCTION: TCM_Return_Type tcmNddsResponderInit (CSFsm fsm)
 *
 * DESCRIPTION: Need to invoke this within CS at startup, to initialize all 
 *              the things needed for CS to respond to TCM commands via NDDS.
 *
 *******************************************************************/
TCM_Return_Type tcmNddsResponderInit (CSFsm fsm)
{
  _fsm = fsm;
  NDDSProducerPropertiesStorage producerProperties;
  NDDSConsumerPropertiesStorage consumerProperties;

  Stimulus_Consumer = NddsConsumerCreate("Stimulus-Consumer", 
					 NDDS_POLLED_RELIABLE,
					 1000, 0.0);
  // Handle any problems with reliable messages.
  NddsConsumerPropertiesGet(Stimulus_Consumer, &consumerProperties);
  consumerProperties.receiveQueueSize = NDDS_RELIABLE_QUEUE_SIZE;
  consumerProperties.reliableStatusRtn = consumerReliableStatusFn;
  NddsConsumerPropertiesSet(Stimulus_Consumer, &consumerProperties);

  tcmNddsStimulusNddsRegister();
  NddsConsumerSubscriptionAdd(Stimulus_Consumer,
			      "tcmNddsStimulus", "tcmStimulus",
			      &tcmNddsStimulusVar, tcmNddsStimulusHandler,
			      NULL);

  Response_Producer = NddsProducerCreate("Response-Producer", 
					 NDDS_ASYNCHRONOUS, 15.0, 1.0);
  // Handle any problems with reliable messages.
  NddsProducerPropertiesGet(Response_Producer, &producerProperties);
  producerProperties.sendQueueSize = NDDS_RELIABLE_QUEUE_SIZE;
  producerProperties.reliableStatusRtn = producerReliableStatusFn;
  NddsProducerPropertiesSet(Response_Producer, &producerProperties);

  tcmNddsResponseNddsRegister();
  NddsProducerProductionAdd(Response_Producer,
			    "tcmNddsResponse", "tcmResponse",
			    &tcmNddsResponseVar, 0, 
			    NULL, NULL);

  return (Stimulus_Consumer != NULL && Response_Producer != NULL
	  ? TCM_Ok : TCM_Error);
}

/*******************************************************************
 *
 * FUNCTION: void tcmNddsResponderPoll (void)
 *
 * DESCRIPTION: This is invoked at the CS side, to handle stimulus 
 *              messages from TCM.  It should be called periodically.
 *
 *******************************************************************/
void tcmNddsResponderPoll (void)
{
  if (Stimulus_Consumer) NddsConsumerPoll(Stimulus_Consumer);
}

/*******************************************************************
 *
 * FUNCTION: void tcmNddsRespond (int refId, String response)
 *
 * DESCRIPTION: This is invoked at the CS side, to send a response back to TCM.
 *              The "refId" is associated with a task tree node, and the
 *              response is either SUCCESS_RESPONSE, or an arbitrary string
 *              denoting failure.
 *
 *******************************************************************/
void tcmNddsRespond (int refId, String response)
{
#ifdef TRACE_RESPONDER
    fprintf(stderr, "tcmNddsRespond: %d %s\n", refId, response);
#endif
  tcmNddsResponseVar.id = refId;
  tcmNddsResponseVar.response = (char *)response;

  NddsProducerSample(Response_Producer);
}

/*******************************************************************
 *
 * FUNCTION: TCM_Return_Type tcmNddsSenderInit (void)
 *
 * DESCRIPTION: Need to invoke this within the TCM process at startup,
 *              to initialize all the things needed for TCM to command
 *              CS via NDDS.
 *
 *******************************************************************/
TCM_Return_Type tcmNddsSenderInit (void)
{
  NDDSProducerPropertiesStorage producerProperties;
  NDDSConsumerPropertiesStorage consumerProperties;

  Response_Consumer = NddsConsumerCreate("Response-Consumer", 
					 NDDS_POLLED_RELIABLE,
					 1000, 0.0);
  // Handle any problems with reliable messages.
  NddsConsumerPropertiesGet(Response_Consumer, &consumerProperties);
  consumerProperties.receiveQueueSize = NDDS_RELIABLE_QUEUE_SIZE;
  consumerProperties.reliableStatusRtn = consumerReliableStatusFn;
  NddsConsumerPropertiesSet(Response_Consumer, &consumerProperties);

  tcmNddsResponseNddsRegister();
  NddsConsumerSubscriptionAdd(Response_Consumer,
			      "tcmNddsResponse", "tcmResponse",
			      &tcmNddsResponseVar, tcmNddsResponseHandler,
			      NULL);

  Stimulus_Producer = NddsProducerCreate("Stimulus-Producer", 
					 NDDS_ASYNCHRONOUS, 15.0, 1.0);
  // Handle any problems with reliable messages.
  NddsProducerPropertiesGet(Stimulus_Producer, &producerProperties);
  producerProperties.sendQueueSize = NDDS_RELIABLE_QUEUE_SIZE;
  producerProperties.reliableStatusRtn = producerReliableStatusFn;
  NddsProducerPropertiesSet(Stimulus_Producer, &producerProperties);

  tcmNddsStimulusNddsRegister();
  NddsProducerProductionAdd(Stimulus_Producer,
			    "tcmNddsStimulus", "tcmStimulus",
			    &tcmNddsStimulusVar, 0,
			    NULL, NULL);

  return (Response_Consumer != NULL && Stimulus_Producer != NULL
	  ? TCM_Ok : TCM_Error);
}

/*******************************************************************
 *
 * FUNCTION: void tcmNddsSenderPoll (void)
 *
 * DESCRIPTION: This is invoked at the TCM side, to handle response 
 *              messages from CS.  It should be called periodically.
 *
 *******************************************************************/
void tcmNddsSenderPoll (void)
{
  NddsConsumerPoll(Response_Consumer);
}

/**************************************************************************
 *
 * FUNCTION: TCM_Return_Type tcmNddsSendStimulus (TCM_Task_Tree_Ref parentNode,
 *				                  String stimulus,
 *				                  int numResponses,
 *                                             TCM_CS_Response_Type *responses)
 *
 * DESCRIPTION: Send a stimulus to CS via NDDS and tell CS to set up the
 *              intention to await response(s) to that stimulus (which will
 *              then be sent back via NDDS).
 *  
 ****************************************************************/
TCM_Return_Type tcmNddsSendStimulus (TCM_Task_Tree_Ref parentNode,
				     String stimulus, int numResponses,
				     TCM_CS_Response_Type *responses)
{
  int i;

  if (numResponses > MAX_RESPONSES) {
    fprintf(stderr, 
	    "tcmNddsSendStimulus: Too many responses (%d) for %s sent from %s\n",
	    numResponses, stimulus, TCM_NodeName(parentNode));
    return TCM_Error;
  } else {
    tcmNddsStimulusVar.stimulus = (char *)stimulus;
    tcmNddsStimulusVar.id = TCM_Remember(parentNode);
    tcmNddsStimulusVar.numResponses = numResponses;
    for (i=0; i<numResponses; i++) {
      tcmNddsStimulusVar.responses[i].context = (char *)responses[i].context;
      tcmNddsStimulusVar.responses[i].returnCode = responses[i].returnCode;
      tcmNddsStimulusVar.responses[i].responseString 
	= (char *)responses[i].responseString;
    }
    for (i=numResponses; i<MAX_RESPONSES; i++) {
      tcmNddsStimulusVar.responses[i].context = "";
      tcmNddsStimulusVar.responses[i].returnCode = 0;
      tcmNddsStimulusVar.responses[i].responseString = "";
    }

#ifdef TRACE_RESPONDER
    fprintf(stderr, "tcmNddsSendStimulus: %s (%d) \n", 
	    tcmNddsStimulusVar.stimulus, tcmNddsStimulusVar.id);
    for (i=0; i<numResponses; i++) {
      fprintf(stderr, "  %s %d => %s\n", 
	      tcmNddsStimulusVar.responses[i].context,
	      tcmNddsStimulusVar.responses[i].returnCode,
	      tcmNddsStimulusVar.responses[i].responseString);
    }
#endif

    NddsProducerSample(Stimulus_Producer);

    return TCM_Ok;
 }
}

/**************************************************************************
 *
 * FUNCTION: TCM_Return_Type tcmNddsSendStimulus (TCM_Task_Tree_Ref parentNode,
 *				                  String stimulus,
 *				                  ...);
 *
 * DESCRIPTION: Send a stimulus to CS via NDDS and tell CS to set up the
 *              intention to await response(s) to that stimulus (which will
 *              then be sent back via NDDS).
 *  
 * NOTES: The variable arguments are triples of the form:
 *    <context, retcode, response-string>, where "context" is the name of the
 *    state (a string) in which the response is enabled, "retcode" is the
 *    transition return code that triggers this response, and "response-string"
 *    is what to send back to TCM (either SUCCESS_RESPONSE or an arbitrary
 *    string, signifying failure).  The end of the list is indicated by the
 *    special symbol END_RESPONSES (defined in tcmResponder.h).
 *  
 * eg, tcmNddsSendStimulus(node, "StartBB", "StartTraj", 1, "NoTrajectory",
 *                         "Completed", 0, SUCCESS_RESPONSE, END_RESPONSES);
 *
 **************************************************************************/

TCM_Return_Type tcmNddsSendStimulus (TCM_Task_Tree_Ref parentNode,
				     String stimulus, ...)
{
  int numArgs, numResponses;
  TCM_CS_Response_Type responses[MAX_RESPONSES];
  va_list args;
  
  // Count how many responses there are.
  va_start(args, stimulus);
  for (numArgs = 0; va_arg(args, void *) != END_RESPONSES; numArgs++) {};
  va_end(args);

  numResponses = numArgs/3;
  if (numArgs == 0) {
    fprintf(stderr, "tcmNddsSendStimulus: No responses for %s sent from %s\n",
	    stimulus, TCM_NodeName(parentNode));
    return TCM_Error;
  } else if (numArgs%3 != 0) {
    fprintf(stderr, 
	    "tcmNddsSendStimulus: Uneven response triples for %s sent from %s\n",
	    stimulus, TCM_NodeName(parentNode));
    return TCM_Error;
  } else if (numResponses > MAX_RESPONSES) {
    fprintf(stderr, 
	    "tcmNddsSendStimulus: Too many responses (%d) for %s sent from %s\n",
	    numResponses, stimulus, TCM_NodeName(parentNode));
    return TCM_Error;
  } else {
    va_start(args, stimulus);
    for (int i=0; i < numResponses; i++) {
      responses[i].set(va_arg(args, String), va_arg(args, int),
		       va_arg(args, String));
    }
    va_end(args);

    return tcmNddsSendStimulus(parentNode, stimulus, numResponses, responses);
  }
}
