
#ifndef _TDL_AllocationFunction_H
#define _TDL_AllocationFunction_H

#include "iostream.h"
#include "stdtypes.H"


/*
 * This class encapsulates our TDL Standard Allocation Function signatures...
 * (Along with standard tests in invocation proceedures.)
 */

typedef TCM_Task_Tree_Ref (* _TDL_AllocationFunctionType_Local)
			  ( STRING );
typedef TCM_Task_Tree_Ref (* _TDL_AllocationFunctionType_Either)
			  ( STRING, STRING, BOOLEAN );
typedef TCM_Task_Tree_Ref (* _TDL_AllocationFunctionType_Distributed)
			  ( STRING, STRING );

class _TDL_AllocationFunction
{
public:
  enum TYPES
  {
    LOCAL         = 0,  /* Original, non-distributed. */
    EITHER        = 1,  /* Either non-distributed or distributed */
    DISTRIBUTED   = 2,  /* ONLY distributed. */

    NULL_FUNCTION = -1,
    UNKNOWN       = -2  /* We are having a bad day... */
  };

  static const char * getStringForType (_TDL_AllocationFunction::TYPES theType)
  {
    switch ( theType )
    {
      case _TDL_AllocationFunction::LOCAL:         return "LOCAL";
      case _TDL_AllocationFunction::EITHER:        return "EITHER";
      case _TDL_AllocationFunction::DISTRIBUTED:   return "DISTRIBUTED";
      case _TDL_AllocationFunction::NULL_FUNCTION: return "NULL";
      case _TDL_AllocationFunction::UNKNOWN:       return "UKNOWN";
      default:                                     return "INVALID_VALUE";
    }
  }


protected:
  union
  {
    _TDL_AllocationFunctionType_Local        local;
    _TDL_AllocationFunctionType_Either       either;
    _TDL_AllocationFunctionType_Distributed  distributed;
  } function;

  _TDL_AllocationFunction::TYPES  type;


public:
  _TDL_AllocationFunction ()
    : type ( _TDL_AllocationFunction::NULL_FUNCTION )
  { function.local =  _TDL_AllocationFunctionType_Local ( NULL ); }

  _TDL_AllocationFunction ( _TDL_AllocationFunctionType_Local  theLocal )
    : type ( _TDL_AllocationFunction::LOCAL )
  { function.local = theLocal; }

  _TDL_AllocationFunction ( _TDL_AllocationFunctionType_Either  theEither )
    : type ( _TDL_AllocationFunction::EITHER )
  { function.either = theEither; }

  _TDL_AllocationFunction (
		      _TDL_AllocationFunctionType_Distributed  theDistributed )
    : type ( _TDL_AllocationFunction::DISTRIBUTED )
  { function.distributed = theDistributed; }

  _TDL_AllocationFunction (
			const _TDL_AllocationFunction & theAllocationFunction )
    : type ( theAllocationFunction.type )
  { function = theAllocationFunction.function; }

  _TDL_AllocationFunction & operator = (
			const _TDL_AllocationFunction & theAllocationFunction )
  {
    function = theAllocationFunction.function;
    type     = theAllocationFunction.type;
    return * this;
  }


  ~_TDL_AllocationFunction()
  {
    function.local = _TDL_AllocationFunctionType_Local ( NULL );
    type           = _TDL_AllocationFunction::UNKNOWN;
  }



  _TDL_AllocationFunction::TYPES getType() const { return type; }

  const char * getTypeString() const { return getStringForType ( getType() ); }



  void * getVoidFunctionPointer() const
  {
    switch ( getType() )
    {
      case _TDL_AllocationFunction::LOCAL:
	return (void *) (function.local);

      case _TDL_AllocationFunction::EITHER:
	return (void *) (function.either);

      case _TDL_AllocationFunction::DISTRIBUTED:
	return (void *) (function.distributed);

      default:
	return (void *) NULL;
    }
  }


  _TDL_AllocationFunctionType_Local  getLocalFunction() const
  {
    if ( getType() == _TDL_AllocationFunction::LOCAL )
      return function.local;
    else
    {
      TDL::getLogStream()
	<< "[_TDL_AllocationFunction:getLocalFunction]  Error:  "
	<< "Tried to access LOCAL function.  Actual function type is "
	<< getTypeString() << endl;
      return _TDL_AllocationFunctionType_Local ( NULL );
    }
  }

  _TDL_AllocationFunctionType_Either  getEitherFunction() const
  {
    if ( getType() == _TDL_AllocationFunction::EITHER )
      return function.either;
    else
    {
      TDL::getLogStream()
	<< "[_TDL_AllocationFunction:getEitherFunction]  Error:  "
	<< "Tried to access EITHER function.  Actual function type is "
	<< getTypeString() << endl;
      return _TDL_AllocationFunctionType_Either ( NULL );
    }
  }


  _TDL_AllocationFunctionType_Distributed  getDistributedFunction() const
  {
    if ( getType() == _TDL_AllocationFunction::DISTRIBUTED )
      return function.distributed;
    else
    {
      TDL::getLogStream()
	<< "[_TDL_AllocationFunction:getDistributedFunction]  Error:  "
	<< "Tried to access DISTRIBUTED function.  Actual function type is "
	<< getTypeString() << endl;
      return _TDL_AllocationFunctionType_Distributed ( NULL );
    }
  }


  BOOLEAN isNull() const
  {
    switch ( getType() )
    {
      case _TDL_AllocationFunction::LOCAL:
	return TO_BOOLEAN (    function.local
			    == _TDL_AllocationFunctionType_Local(NULL) );

      case _TDL_AllocationFunction::EITHER:
	return TO_BOOLEAN (    function.either
			    == _TDL_AllocationFunctionType_Either(NULL) );

      case _TDL_AllocationFunction::DISTRIBUTED:
	return TO_BOOLEAN (    function.distributed
			    == _TDL_AllocationFunctionType_Distributed(NULL) );

	/* All other cases... */
      default:
	return TRUE;
    }
  }


  BOOLEAN isDistributable() const
  {
    return
      TO_BOOLEAN (   ( getType() == _TDL_AllocationFunction::EITHER      )
		  || ( getType() == _TDL_AllocationFunction::DISTRIBUTED ) );
  }



  TCM_Task_Tree_Ref invokeFunction ( STRING theTaskName  = STRING(NULL),
				     STRING theAgentName = STRING(NULL) ) const
  {
    switch ( getType() )
    {
      case _TDL_AllocationFunction::LOCAL:
	return function.local ( theTaskName );

      case _TDL_AllocationFunction::EITHER:
	    /* Third argument is "theIsDistributedRemoteSide = FALSE" */
	return function.either ( theTaskName, theAgentName, FALSE );

      case _TDL_AllocationFunction::DISTRIBUTED:
	if ( theAgentName !=  STRING(NULL) )
	  return function.distributed ( theTaskName, theAgentName );
	else
	{
	  TDL::getLogStream()
	    << "[_TDL_AllocationFunction:invokeFunction]  Error:  "
	    << "Trying to invoke a DISTRIBUTED function with a NULL "
	    << "agent-name.  Aborting.  Returning NULL TCM_Task_Tree_Ref."
	    << endl;

	  return TCM_Task_Tree_Ref(); /* Return a temporary NULL value */
	}

      default:
	TDL::getLogStream()
	  << "[_TDL_AllocationFunction:invokeFunction]  Error:  "
	  << "Unknown function type (" << getTypeString()
	  << ").  Aborting.  Returning NULL TCM_Task_Tree_Ref." << endl;

	return TCM_Task_Tree_Ref(); /* Return a temporary NULL value */
    }
  }

};

#endif /* _TDL_AllocationFunction_H */
